use std::fmt::Display;

/// The error that is returned by the [`SchemaInspector`] when the provided JSON to the `inspect` function
/// fails to conform to the schema the inspector was compiled with.
///
/// The error type internally holds a context containing detailed information about what part of the JSON differs from the schema
#[derive(Debug, Clone, PartialEq, Eq)]
pub struct SchemaError {
    pub(crate) context: Vec<ErrContext>,
}

/// The interal type containing metadata where the error has occurred along with why
#[derive(Debug, Clone, PartialEq, Eq)]
#[cfg_attr(feature = "serde", derive(serde::Serialize))]
pub struct ErrContext {
    /// the field where the validation failed
    pub field: String,
    /// detailed description of why the validation failed
    pub description: String,
}

impl SchemaError {
    /// Get the context of the error
    pub fn context(&self) -> &[ErrContext] {
        &self.context
    }
}

impl std::error::Error for SchemaError {}

impl Display for SchemaError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let num_errors = self.context.len();
        let noun = if num_errors < 2 { "error" } else { "errors" };
        write!(
            f,
            "JSON doesn't match expected schema: found {num_errors} {noun}",
        )
    }
}

/// All the possible errors that can occur while initializing [`SchemaInspector`]
#[derive(Debug, Clone, PartialEq, Eq)]
pub enum InitError {
    SchemaNotFound(String),
    ResolutionFailure(String), // contains the reference that was not resolved
    InvalidSchema(Option<String>), // contains the missing field in the schema
    CompileFailure,
    InferenceFailure,
}

impl std::error::Error for InitError {}

impl Display for InitError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let base = "failed to initialize Inspector:";
        match self {
            Self::SchemaNotFound(schema) => {
                write!(f, "{base} {schema} was not found in the provided schemas")
            }
            Self::ResolutionFailure(schema) => write!(
                f,
                "{base} failed to resolve references: referenced schema {schema} was not found"
            ),
            Self::InferenceFailure => write!(f, "{base} name of the type couldn't be inferred"),
            Self::CompileFailure => write!(f, "{base} failed to compile schema"),
            Self::InvalidSchema(schema) => write!(
                f,
                "{base} invalid schema provided: {} is missing or is not an object",
                schema.as_deref().unwrap_or("field")
            ),
        }
    }
}
